import json
import os
from typing import Optional

from sqlalchemy import create_engine
from sqlalchemy.orm import Session, scoped_session, sessionmaker

from hojo.config import Config


class ConnectionCredentialError(RuntimeError):
    pass


class Connection:
    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(Connection, cls).__new__(cls)
        return cls._instance

    def __init__(self) -> None:
        self._session = None
        self._engine = None
        self.db_uri = Config.get("db_uri") or os.environ.get("DB_URI")

        if not self.db_uri:
            raise ConnectionCredentialError("Invalid database credentials.")

    @property
    def session(self) -> Session:
        if self._session:
            return self._session

        self._session = self.create_session()
        return self._session

    def create_session(self) -> Session:
        session_factory = sessionmaker(
            bind=self._get_engine(), future=True, expire_on_commit=False
        )
        session: Session = scoped_session(session_factory)  # type: ignore
        return session

    @classmethod
    def reset(cls):
        cls._instance = None

    def _get_engine(self):
        if self._engine:
            return self._engine

        self._engine = create_engine(self.db_uri)  # type: ignore

        return self._engine



from __future__ import annotations

import json
from typing import List, Optional, Union

from attrs import define, make_class

from hojo.converter import SchemaConverter


class BaseSchema:
    @classmethod
    def load(cls, data: Union[dict, BaseSchema], **kwargs):
        if isinstance(data, BaseSchema):
            data = data.dump()

        return SchemaConverter.structure(data, cls)

    def dump(
        self,
        skip_none: bool = False,
        only: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None,
    ):
        data = SchemaConverter.unstructure(self)

        if only:
            data = {key: val for key, val in data.items() if key in only}
        elif exclude:
            data = {key: val for key, val in data.items() if key not in exclude}

        if skip_none:
            data = {key: val for key, val in data.items() if val is not None}

        return data

    def dumps(self, **kwargs):
        data = self.dump(**kwargs)

        return json.dumps(data)


def schema(cls) -> BaseSchema:
    klass = define(cls)

    class_name = cls.__name__
    ProxySchema = make_class(class_name, {}, bases=(klass, BaseSchema), cmp=False)

    ProxySchema.__name__ = cls.__name__
    ProxySchema.__qualname__ = cls.__qualname__
    ProxySchema.__module__ = cls.__module__
    ProxySchema.__doc__ = cls.__doc__

    return ProxySchema



class Config:
    _instance = None
    _configurations = {}

    @classmethod
    def instance(cls):
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    @classmethod
    def set(cls, key, value):
        cls.instance()._configurations[key] = value

    @classmethod
    def get(cls, key):
        return cls.instance()._configurations.get(key)

    @classmethod
    def reset(cls):
        cls._instance = None
        cls._configurations = {}



from enum import EnumType
from typing import Union, get_args, get_origin

from attrs import fields, resolve_types
from pluralizer import Pluralizer
from sqlalchemy import Column, Index, Table
from sqlalchemy.dialects.postgresql import UUID as SQLAlchemyUUID
from sqlalchemy.orm import registry
from sqlalchemy.types import Boolean, Date, DateTime, Float, Integer, String
from sqlalchemy_utils.types import EnrichedDateTimeType, EnrichedDateType

from hojo.base import BaseModel

MAPPER_REGISTRY = registry()


def automap(registry=None):
    registry = registry or MAPPER_REGISTRY

    for model in BaseModel._registry:
        builder = TableBuilder(registry, model)
        builder.automap()

    return registry


class TypeTranslator:
    def __init__(self, model):
        self.model = model
        self.type_mapping = {
            "int": Integer,
            "float": Float,
            "str": String,
            "bool": Boolean,
            "datetime": DateTime,
            "date": Date,
            "UUID": SQLAlchemyUUID(as_uuid=True),
        }

        resolve_types(self.model)

    def translate(self, field):
        field_type = field.type
        origin_type = get_origin(field_type)
        field_args = get_args(field_type)

        if origin_type is Union and len(field_args) == 2 and type(None) in field_args:
            field_type = next(t for t in field_args if t is not type(None))
            field_type_name = field_type.__name__
        else:
            field_type_name = (
                field_type if type(field_type) == str else field_type.__name__
            )

        if field_type_name in self.type_mapping:
            return self.type_mapping[field_type_name]
        if field_type_name == "DateTime":
            return EnrichedDateTimeType()
        if field_type_name == "Date":
            return EnrichedDateType()
        elif isinstance(field_type, EnumType):
            return String
        else:
            raise TypeError(
                f"Unsupported type: {field_type} for {self.model.__name__}.{field_type_name}'"
            )


class RelationshipManager:
    def __init__(self):
        self.relationships = {}

    def add_relationship(self, field_name, field_info):
        pass

    def get_relationships(self):
        return self.relationships


class TableBuilder:
    def __init__(self, mapper_registry, model):
        self.mapper_registry = mapper_registry
        self.model = model
        self.translator = TypeTranslator(model)
        self.relationship_manager = RelationshipManager()

    def build_table(self):
        columns = []
        uniques = {}
        indexes = {}

        for field_info in fields(self.model):
            if self._is_relationship_field(field_info):
                self.relationship_manager.add_relationship(field_info.name, field_info)
                continue

            column_type = self.translator.translate(field_info)

            is_primary_key = field_info.name == "id"  # id is the primary key by default
            is_nullable = field_info.default is None

            is_index = self.get_constraint("index", indexes, field_info)
            is_unique = self.get_constraint("unique", uniques, field_info)

            column = Column(
                field_info.name,
                column_type,
                primary_key=is_primary_key,
                nullable=is_nullable,
                index=is_index,
                unique=is_unique,
            )

            columns.append(column)

        for index in indexes:
            idx_fields = sorted(indexes[index])
            index_name = "ix_" + "_".join(idx_fields)
            columns.append(Index(index_name, *idx_fields))

        table = Table(
            Pluralizer().plural((self.model.__name__.lower())),
            self.mapper_registry.metadata,
            *columns,
        )

        return table

    def get_constraint(self, contraint_type, contraints, field_info):
        idx = field_info.metadata.get(contraint_type)
        match idx:
            case bool():
                is_index = idx
            case str():
                is_index = False
                contraints[idx] = contraints.get(idx) or []
                contraints[idx].append(field_info.name)
            case _:
                is_index = False
        return is_index

    def _is_relationship_field(self, field_info):
        info = getattr(field_info, "metadata") or {}
        return info.get("has_many") or info.get("belongs_to") or info.get("has_one")

    def automap(self):
        table = self.build_table()

        self.mapper_registry.map_imperatively(
            self.model,
            table,
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, List, Tuple, Type, TypeVar, Union

from sqlalchemy import delete, select, update
from sqlalchemy.exc import MultipleResultsFound, NoResultFound
from sqlalchemy.orm import Session

T = TypeVar("T")


@dataclass
class LookupFilter:
    field_name: str
    lookup: str
    value: Any
    exclude: bool


class QuerySet:
    def __init__(self, model_class: Type[T], session: Session) -> None:
        self.model_class = model_class
        self.session = session
        self.query = select(model_class)
        self._result_cache = None
        self._executed = False
        self.lookup_filters: List[LookupFilter] = []

    def __iter__(self):
        if not self._executed:
            self._select()
        return iter(self._result_cache)

    def __len__(self):
        if not self._executed:
            self._select()
        return len(self._result_cache)

    def __bool__(self):
        if not self._executed:
            self._select()
        return bool(self._result_cache)

    def __repr__(self):
        if not self._executed:
            return "<QuerySet [not executed]>"
        return repr(self._result_cache)

    def _clone(self) -> QuerySet[T]:
        qs = self.__class__(self.model_class, self.session)
        qs.query = self.query
        qs.lookup_filters = self.lookup_filters.copy()
        return qs

    def compile_conditions(self):
        if self.lookup_filters:
            conditions = []
            for lookup in self.lookup_filters:
                column = getattr(self.model_class, lookup.field_name)
                lookup_name = lookup.lookup
                if lookup_name == "eq":
                    filter_expr = column == lookup.value
                elif lookup_name == "gt":
                    filter_expr = column > lookup.value
                elif lookup_name == "gte":
                    filter_expr = column >= lookup.value
                elif lookup_name == "lt":
                    filter_expr = column < lookup.value
                elif lookup_name == "lte":
                    filter_expr = column <= lookup.value
                elif lookup_name == "in":
                    filter_expr = column.in_(lookup.value)
                elif lookup_name == "isnull":
                    filter_expr = column.is_(None)
                elif lookup_name == "between":
                    filter_expr = column.between(*lookup.value)
                else:
                    filter_expr = getattr(column, lookup_name)(lookup.value)
                if lookup.exclude:
                    filter_expr = ~filter_expr
                conditions.append(filter_expr)

            return conditions

    def _mount_filters(self):
        conditions = self.compile_conditions()
        self.query = self.query.where(*conditions)

    def _select(self):
        self._mount_filters()

        self._result_cache = self.session.execute(self.query).scalars().all()
        self._executed = True

    def _apply_filter(self, exclude=False, **kwargs) -> QuerySet[T]:
        cloned_qs = self._clone()

        lookups = [
            "like",
            "ilike",
            "contains",
            "startswith",
            "endswith",
            "gt",
            "lt",
            "gte",
            "lte",
            "isnull",
            "in",
            "between",
        ]

        for key, val in kwargs.items():
            parts = key.split("__")
            field_name = parts[0]
            if len(parts) > 1:
                lookup = parts[1]
                if lookup not in lookups:
                    raise ValueError(f"Unsupported lookup filter: {lookup}")
            else:
                lookup = "eq"

            cloned_qs.lookup_filters.append(
                LookupFilter(field_name, lookup, val, exclude)
            )

        return cloned_qs

    def all(self) -> QuerySet[T]:
        return self._clone()

    def first(self) -> Union[T, None]:
        if not self._executed:
            self._select()
        return self._result_cache[0] if self._result_cache else None

    def last(self) -> Union[T, None]:
        if not self._executed:
            self._select()
        return self._result_cache[-1] if self._result_cache else None

    def order_by(self, *fields: str) -> QuerySet[T]:
        cloned_qs = self._clone()
        for field in fields:
            cloned_qs.query = cloned_qs.query.order_by(getattr(self.model_class, field))
        return cloned_qs

    def filter(self, **kwargs) -> QuerySet[T]:
        return self._apply_filter(**kwargs)

    def exclude(self, **kwargs) -> QuerySet[T]:
        return self._apply_filter(exclude=True, **kwargs)

    def create(self, **kwargs) -> T:
        obj = self.model_class.load(kwargs)
        self.session.add(obj)
        self.session.commit()
        return obj

    def delete(self, **kwargs) -> None:
        conditions = self.compile_conditions()
        delete_query = delete(self.model_class).where(*conditions)

        self.session.execute(delete_query)
        self.session.commit()

    def update(self, **kwargs) -> None:
        conditions = self.compile_conditions()
        update_query = update(self.model_class).where(*conditions).values(**kwargs)

        self.session.execute(update_query)
        self.session.commit()

    def get(self, **kwargs) -> T:
        try:
            return self.filter(**kwargs).first()
        except NoResultFound:
            raise ValueError("No results found.")
        except MultipleResultsFound:
            raise ValueError("Multiple results returned for `get`.")



from typing import Any, ClassVar, List, Optional, Type
from uuid import UUID

from attrs import define
from attrs import field as attrfield
from attrs import make_class
from pendulum import DateTime, now
from uuid6 import uuid7

from hojo.orm.manager import Manager, ModelDescriptor
from hojo.schema import BaseSchema


class BaseModel(BaseSchema):
    _registry: List["BaseModel"] = []

    objects: ClassVar = ModelDescriptor(Manager())


def field(
    primary_key: bool = False,
    index: bool = False,
    unique: bool = False,
    has_many: Optional[BaseModel] = None,
    belongs_to: Optional[BaseModel] = None,
    has_one: Optional[BaseModel] = None,
    **kwargs,
) -> Any:
    custom_metadata = {
        "primary_key": primary_key,
        "index": index,
        "unique": unique,
        "has_many": has_many,
        "belongs_to": belongs_to,
        "has_one": has_one,
    }

    # If metadata already exists in kwargs, update it; otherwise, create it
    if "metadata" in kwargs:
        kwargs["metadata"].update(custom_metadata)
    else:
        kwargs["metadata"] = custom_metadata

    return attrfield(**kwargs)


def model(cls, slots=False) -> Type[BaseModel]:
    klass = define(cls)

    class_fields = {
        "id": attrfield(type=UUID, factory=uuid7, kw_only=True),
        "created_at": attrfield(
            type=DateTime, factory=lambda: now("UTC"), kw_only=True
        ),  # type: ignore
        "updated_at": attrfield(
            type=DateTime, factory=lambda: now("UTC"), kw_only=True
        ),  # type: ignore
    }

    class_name = cls.__name__
    ProxyModel = make_class(
        class_name, class_fields, bases=(klass, BaseModel), cmp=False, slots=slots
    )

    ProxyModel.__name__ = cls.__name__
    ProxyModel.__qualname__ = cls.__qualname__
    ProxyModel.__module__ = cls.__module__
    ProxyModel.__doc__ = cls.__doc__

    BaseModel._registry.append(ProxyModel)

    return ProxyModel



from datetime import date, datetime
from enum import IntEnum, StrEnum
from uuid import UUID

import pendulum
from cattrs import Converter

SchemaConverter = Converter()

# UUID converter
SchemaConverter.register_unstructure_hook(UUID, lambda id: str(id))
SchemaConverter.register_structure_hook(UUID, lambda id, _: UUID(id))


# StrEnum converter
def _unstructure_strenum(enum_value: StrEnum | str):
    if type(enum_value) == str:
        return enum_value
    return enum_value.value


def _structure_strenum(value, enum_type):
    return enum_type(value)


SchemaConverter.register_unstructure_hook(StrEnum, _unstructure_strenum)
SchemaConverter.register_structure_hook(StrEnum, _structure_strenum)


# IntEnum converter
def _unstructure_strenum(enum_value: IntEnum | int):
    if type(enum_value) == int:
        return enum_value
    return enum_value.value


def _structure_strenum(value, enum_type):
    return enum_type(value)


SchemaConverter.register_unstructure_hook(IntEnum, _unstructure_strenum)
SchemaConverter.register_structure_hook(IntEnum, _structure_strenum)

# pendulum.Date converter
SchemaConverter.register_unstructure_hook(pendulum.Date, lambda dt: dt.to_date_string())
SchemaConverter.register_structure_hook(
    pendulum.Date, lambda dt, _: pendulum.parse(dt).date()
)


# pendulum.DateTime converter
def _unstructure_pendulum_datetime(pdt: pendulum.DateTime | datetime):
    if type(pdt) == datetime:
        return pendulum.instance(pdt).to_iso8601_string()

    return pdt.to_iso8601_string()  # type: ignore


def _structure_pendulum_datetime(value, _):
    return pendulum.parse(value)


SchemaConverter.register_unstructure_hook(
    pendulum.DateTime, _unstructure_pendulum_datetime
)

SchemaConverter.register_structure_hook(pendulum.DateTime, _structure_pendulum_datetime)



from __future__ import annotations

from typing import Any, Optional, TypeVar

from hojo.connection import Connection
from hojo.orm.queryset import QuerySet
from hojo.schema import BaseSchema

T = TypeVar("T")


class Manager:
    """
    Interface for database operations that provides a high-level, abstracted set of
    database-querying methods.
    """

    def __init__(self) -> None:
        self._model_class: Optional[type[T]] = None

    @property
    def model_class(self) -> type[T]:
        if self._model_class is None:
            raise ValueError("Model class is not set for this manager.")
        return self._model_class

    @model_class.setter
    def model_class(self, value: type[T]) -> None:
        self._model_class = value

    def get_queryset(self) -> QuerySet[T]:
        return QuerySet(self._model_class, Connection().session)

    def all(self) -> QuerySet[T]:
        return self.get_queryset().all()

    def filter(self, **kwargs) -> QuerySet[T]:
        return self.get_queryset().filter(**kwargs)

    def exclude(self, **kwargs) -> QuerySet[T]:
        return self.get_queryset().exclude(**kwargs)

    def first(self) -> T:
        return self.get_queryset().first()

    def last(self) -> T:
        return self.get_queryset().last()

    def create(self, **data) -> T:
        if isinstance(data, BaseSchema):
            data = data.dump()

        return self.get_queryset().create(**data)

    def get(self, **kwargs) -> T:
        return self.get_queryset().get(**kwargs)

    def get_or_create(self, **kwargs) -> T:
        return self.get_queryset().get_or_create(**kwargs)

    def bulk_create(self, objs):
        return self.get_queryset().bulk_create(objs)

    def bulk_update(self, objs):
        return self.get_queryset().bulk_update(objs)

    def count(self) -> int:
        return self.get_queryset().count()

    def order_by(self, *fields: str) -> QuerySet[T]:
        return self.get_queryset().order_by(*fields)

    def values(self, *fields: str) -> list[dict]:
        return self.get_queryset().values(*fields)

    def values_list(self, *fields: str, flat=False) -> list:
        return self.get_queryset().values_list(*fields, flat=flat)

    def exists(self) -> bool:
        return self.get_queryset().exists()

    def delete(self) -> None:
        self.get_queryset().delete()

    def update(self, **kwargs) -> None:
        self.get_queryset().update(**kwargs)

    def annotate(self, **kwargs) -> QuerySet[T]:
        return self.get_queryset().annotate(**kwargs)

    def aggregate(self, **kwargs) -> dict:
        return self.get_queryset().aggregate(**kwargs)

    def reverse(self) -> QuerySet[T]:
        return self.get_queryset().reverse()

    def distinct(self) -> QuerySet[T]:
        return self.get_queryset().distinct()


class ModelDescriptor:
    def __init__(self, manager):
        self.manager = manager

    def __get__(self, instance, owner):
        if instance is not None:
            raise AttributeError("Manager isn't accessible via model instances")

        # Bind the model class to the manager and return it
        self.manager.model_class = owner
        return self.manager



from hojo.base import BaseModel, field, model
from hojo.config import Config
from hojo.connection import Connection
from hojo.orm.mapper import automap
from hojo.schema import BaseSchema, schema


class Hojo:
    @staticmethod
    def config(key=None, **configs):
        if key:
            return Config.get(key)

        for key, value in configs.items():
            Config.set(key, value)
